//
//  GENXWriter.m
//
//  Created by Quentin Carnicelli on 10/30/04.
//  Copyright 2004 Rogue Amoeba Software, LLC. All rights reserved.
//
//  Permission is hereby granted, free of charge, to any person obtaining a copy
//  of this software and associated documentation files (the "Software"), to
//  deal in the Software without restriction, including without limitation the
//  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
//  sell copies of the Software, and to permit persons to whom the Software is
//  furnished to do so, subject to the following conditions:
//  
//  The above copyright notice and this permission notice shall be included in
//  all copies or substantial portions of the Software.
//  
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
//  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
//  IN THE SOFTWARE.

#import "GENXWriter.h"

NSString* GENXWriterException = @"GENXWriterException";

@interface GENXWriter (Private)
static genxStatus _GENXWriter_send( void* userData, constUtf8 str );
static genxStatus _GENXWriter_sendBounded( void* userData, constUtf8 start, constUtf8 end );
static genxStatus _GENXWriter_flush( void* userData );
@end

@implementation GENXWriter

+ (NSString*)genxVersion
{
	char* str = genxGetVersion();
	return str != nil ? [NSString stringWithCString: str] : nil;
}

+ (NSString*)prefixOfNamespace: (GENXNamespace)ns
{
	utf8 str = genxGetNamespacePrefix( ns );
	return str != nil ? [NSString stringWithUTF8String: str] : nil;
}

- (void)_raiseError: (genxStatus)status
{
	char* errDesc;
	
	if( status == GENX_SUCCESS )
		return;

	errDesc = genxGetErrorMessage( _writer, status );

	[NSException raise: GENXWriterException format: @"GENXWriter %p: %s (genx status code: #%d)", self, errDesc, status];
}

- (id)_basicInit
{
	self = [super init];
	if( self )
	{
		_writer = genxNew( nil, nil, self );
		if( !_writer )
		{
			[self release];
			return nil;
		}
	}
	
	return self;
}

- (id)initForWritingWithFile: (FILE*)file
{
	self = [self _basicInit];
	if( self )
	{
		genxStatus status;
		
		_file = file;
		if( _file == nil )
		{
			[self release];
			return nil;
		}
		
		status = genxStartDocFile( _writer, _file );
		if( status != GENX_SUCCESS )
		{
			[self release];
			[self _raiseError: status];
			return nil;
		}
	}
	
	return self;
}

- (id)initForWritingWithMutableData:(NSMutableData*)data
{
	self = [self _basicInit];
	if( self )
	{
		genxStatus status;

		_data = [data retain];
		if( _data == nil )
		{
			[self release];
			return nil;
		}
		
		_sender.send			= _GENXWriter_send;
		_sender.sendBounded		= _GENXWriter_sendBounded;
		_sender.flush			= _GENXWriter_flush;
		status = genxStartDocSender( _writer, &_sender );
		if( status != GENX_SUCCESS )
		{
			[self release];
			[self _raiseError: status];
			return nil;
		}
	}
	
	return self;
}

- (void)dealloc
{
	if( _writer )
	{
		genxDispose( _writer );
		_writer = nil;
	}
		
	if( _data )
	{
		[_data release];
		_data = nil;
	}

	[super dealloc];
}

- (NSData*)xmlData
{
	return _data;
}

- (FILE*)xmlFile
{
	return _file;
}

- (void)finishedWriting
{
	genxEndDocument( _writer );
}

#pragma mark -

- (GENXNamespace)declareNamespace: (NSString*)prefix withURI: (NSString*)uri
{
	GENXNamespace ns;
	genxStatus status;
	ns = genxDeclareNamespace( _writer, [uri UTF8String], [prefix UTF8String], &status );
	return ns;
}

- (GENXElement)declareElement: (NSString*)name inNamespace: (GENXNamespace)ns
{
	GENXElement element;
	genxStatus status;
	element = genxDeclareElement( _writer, ns, [name UTF8String], &status );
	return element;
}

- (GENXAttribute)delcareAttribute: (NSString*)name inNamespace: (GENXNamespace)ns
{
	GENXAttribute attribute;
	genxStatus status;
	attribute = genxDeclareAttribute( _writer, ns, [name UTF8String], &status );
	return attribute;
}

- (void)startDeclaredElement: (GENXElement)e
{
	genxStatus status = genxStartElement( e );
	if( status != GENX_SUCCESS )
		[self _raiseError: status ];
}

- (void)startLiteralElement: (NSString*)name xmlns: (NSString*)xmlns
{
	genxStatus status = genxStartElementLiteral( _writer, [xmlns UTF8String], [name UTF8String] );
	if( status != GENX_SUCCESS )
		[self _raiseError: status ];
}

- (void)endElement
{
	genxStatus status = genxEndElement( _writer );
	if( status != GENX_SUCCESS )
		[self _raiseError: status ];
}

- (void)addDeclaredAttribute: (GENXAttribute)attr withValue: (NSString*)val
{
	genxStatus status = genxAddAttribute( attr, [val UTF8String] );
	if( status != GENX_SUCCESS )
		[self _raiseError: status ];
}

- (void)addLiteralAttribute: (NSString*)name withValue: (NSString*)val xmlns: (NSString*)xmlns
{
	genxStatus status = genxAddAttributeLiteral( _writer, [xmlns UTF8String],  [name UTF8String], [val UTF8String] );
	if( status != GENX_SUCCESS )
		[self _raiseError: status ];
}

- (void)addText: (NSString*)str
{
	genxStatus status = genxAddText( _writer, [str UTF8String] );
	if( status != GENX_SUCCESS )
		[self _raiseError: status ];
}

- (void)addCharacter: (unichar)ch
{
	genxStatus status = genxAddCharacter( _writer, ch );
	if( status != GENX_SUCCESS )
		[self _raiseError: status ];
}

- (void)addComment: (NSString*)comment
{
	genxStatus status = genxComment( _writer, [comment UTF8String] );
	if( status != GENX_SUCCESS )
		[self _raiseError: status ];
}

- (void)addProcessingInstruction: (NSString*)target text: (NSString*)text
{
	genxStatus status = genxPI( _writer, [target UTF8String], [text UTF8String] );
	if( status != GENX_SUCCESS )
		[self _raiseError: status ];
}

- (void)addNamespace: (GENXNamespace)ns prefix: (NSString*)prefix
{
	genxStatus status = genxAddNamespace( ns, [prefix UTF8String] );
	if( status != GENX_SUCCESS )
		[self _raiseError: status ];
}

- (void)unsetDefaultNamespace
{
	genxStatus status = genxUnsetDefaultNamespace( _writer );
	if( status != GENX_SUCCESS )
		[self _raiseError: status ];
}


- (BOOL)checkText: (NSString*)str
{
	genxStatus status = genxCheckText( _writer, [str UTF8String] );
	return (status == GENX_SUCCESS);
}

- (NSString*)scrubText: (NSString*)str
{
	const char* inStr;
	char* outStr;
	int changed;
	
	inStr = [str UTF8String];
	outStr = malloc( sizeof( char ) * strlen(inStr) );
	if( !outStr )
		return str;
	
	changed = genxScrubText( _writer, inStr, outStr );
	if( changed )
	{
		str = [NSString stringWithUTF8String: outStr];
	}
	
	free( outStr );
	
	return str;
}

#pragma mark -

static genxStatus _GENXWriter_send( void* userData, constUtf8 str )
{
	GENXWriter* writer = (GENXWriter*)userData;
	if( !writer )
		return GENX_INTERNAL_ERROR;

	[writer->_data appendBytes: str length: strlen(str)];
	
	return GENX_SUCCESS;
}

static genxStatus _GENXWriter_sendBounded( void* userData, constUtf8 start, constUtf8 end )
{
	GENXWriter* writer = (GENXWriter*)userData;
	if( !writer )
		return GENX_INTERNAL_ERROR;

	[writer->_data appendBytes: start length: end-start];
	
	return GENX_SUCCESS;
}

static genxStatus _GENXWriter_flush( void* userData )
{
	return GENX_SUCCESS;
}

@end

#pragma mark -

@implementation GENXWriter (ConvenienceAdditions)

- (void)startLiteralElement: (NSString*)name
{
	[self startLiteralElement: name xmlns: nil];
}

- (void)addLiteralAttribute: (NSString*)name withValue: (NSString*)val
{
	[self addLiteralAttribute: name withValue: val xmlns: nil];
}

- (void)addLiteralElement: (NSString*)name withText: (NSString*)text
{
	[self startLiteralElement: name xmlns: nil];
	[self addText: text];
	[self endElement];
}

@end
